# Copyright 2014 NYBX Inc.
# See LICENSE file for details.

"""
:module: ledgerx.cli
:synopsis: Module responsible for CLI.
:author: Amr Ali <amr@ledgerx.com>
"""

import os
import sys
import logging
import argparse

from ledgerx.version import __version__
from ledgerx.protocol.crypto import KeyPair

def _out(msg, level=logging.INFO):
    _d = {logging.INFO: '*', logging.WARNING: '!', logging.ERROR: 'x'}
    sys.stdout.write("[{0}] {1}\n".format(_d[level], msg))

def _die(msg, success=False):
    if success:
        sys.stdout.write("{0}\n".format(msg))
        sys.exit(0)
    else:
        sys.stderr.write("error: {0}\n".format(msg))
        sys.exit(1)

class CommandLine(object):
    """\
    Parse and logically validate command line arguments.
    """
    _description = "LedgerX Exchange Python Client"
    _epilog = "Report bugs to bugs@ledgerx.com"

    def __init__(self):
        parser = argparse.ArgumentParser(description=self._description,
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=self._epilog)

        # CLI arguments
        parser.add_argument('-g', '--generate-keypair', metavar='FILEPATH',
                default=None,
                help='generate a client key pair and store it in FILEPATH')
        parser.add_argument('--version', action='version',
                version='%(prog)s ' + __version__,
                help='output version information and exit')

        self.parser = parser

    def parse(self):
        """\
        Parse and validate command line arguments and set configurations.
        """
        return self.parser.parse_args()

def ledgerx_main():
    """\
    pyledgerx application entry point.
    """
    # Parse command line arguments
    cmd = CommandLine()
    args = cmd.parse()

    if args.generate_keypair:
        # Extract both the filepath and dirpath
        filepath = args.generate_keypair
        dirpath = os.path.dirname(filepath)

        # Check if dirpath is accessible and writable
        if not os.access(dirpath, os.W_OK | os.X_OK):
            _die('cannot access or write to {0}'.format(dirpath))

        # Generate the key pair certificate file
        _out('generating new key pair')
        KeyPair.generate_certificate(filepath)
        _dir('client key pair has been successfully generated', True)

