# Copyright 2014 NYBX Inc.
# All rights reserved.

"""
:module: ledgerx.commands.v0_0_1
:synopsis: Commands module to issue orders or quotes to the exchange.
:author: Amr Ali <amr@ledgerx.com>

This module corresponds to v0.0.1 protocol messages.
"""

from ledgerx.exceptions import LXError
from ledgerx.mixin import public
from ledgerx.utils import reraise
from ledgerx.context import Context
from ledgerx.api.client import v0_0_1 as api

@public
class Commands(object):
    """\
    An implementation of :class:`ledgerx.commands.Commands`.
    """

    def __init__(self, context):
        self.context = context

    def limit_order(self, **kwargs):
        """\
        Place a limit order on the exchange.

        Not all parameters are applicable to this function.

        :param is_ask: Whether this is an ask or a bid.
        :param size: The order or quote size.
        :param price_in_cents: The price of the contract in cents.
        :param mox2: Should the mox2 mechanism be enabled.
        :param scp_sweeper_mpid: SCP sweeper MPID.
        :param scp_sweeper_cid: SCP sweeper client ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageLimitOrder()
        msg = self.__make_submission(msg, **kwargs)
        return self.__submit(msg)

    def market_order(self, **kwargs):
        """\
        Place a market order on the exchange.

        Not all parameters are applicable to this function.

        :param is_ask: Whether this is an ask or a bid.
        :param size: The order or quote size.
        :param price_in_cents: The price of the contract in cents.
        :param mox2: Should the mox2 mechanism be enabled.
        :param scp_sweeper_mpid: SCP sweeper MPID.
        :param scp_sweeper_cid: SCP sweeper client ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageMarketOrder()
        msg = self.__make_submission(msg, **kwargs)
        return self.__submit(msg)

    def scp_order(self, **kwargs):
        """\
        Place a SCP order on the exchange.

        Not all parameters are applicable to this function.

        :param is_ask: Whether this is an ask or a bid.
        :param size: The order or quote size.
        :param price_in_cents: The price of the contract in cents.
        :param mox2: Should the mox2 mechanism be enabled.
        :param scp_sweeper_mpid: SCP sweeper MPID.
        :param scp_sweeper_cid: SCP sweeper client ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageSCPOrder()
        msg = self.__make_submission(msg, **kwargs)
        return self.__submit(msg)

    def quote(self, **kwargs):
        """\
        Place a quote on the exchange.

        Not all parameters are applicable to this function.

        :param is_ask: Whether this is an ask or a bid.
        :param size: The order or quote size.
        :param price_in_cents: The price of the contract in cents.
        :param mox2: Should the mox2 mechanism be enabled.
        :param scp_sweeper_mpid: SCP sweeper MPID.
        :param scp_sweeper_cid: SCP sweeper client ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageQuote()
        msg = self.__make_submission(msg, **kwargs)
        return self.__submit(msg)

    def cancel_entry(self, contract_id, entry_id):
        """\
        Cancel an entry on a particular contract on the exchange.

        :param contract_id: The contract ID on which the entry resides.
        :param entry_id: The ID of the entry to be canceled.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageCancel()
        msg.contract_id = contract_id
        msg.entry_id = entry_id
        return self.__submit(msg)

    def cancel_replace_entry(self, entry_id, **kwargs):
        """\
        Cancel an existing entry and replace it with another on the exchange.

        :param entry_id: The ID of the entry to be canceled.
        :param is_ask: Whether this is an ask or a bid.
        :param size: The order or quote size.
        :param price_in_cents: The price of the contract in cents.
        :param mox2: Should the mox2 mechanism be enabled.
        :param scp_sweeper_mpid: SCP sweeper MPID.
        :param scp_sweeper_cid: SCP sweeper client ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageCancelReplace()
        msg = self.__make_submission(msg, **kwargs)
        msg.entry_id = entry_id
        return self.__submit(msg)

    def reduce_entry_size(self, contract_id, entry_id, size_decrement):
        """\
        Reduce the size of an existing entry on the exchange.

        :param contract_id: The contract ID on which the entry resides.
        :param entry_id: The ID of the entry to be reduced.
        :param size_decrement: The amount by which the entry's size will be reduced.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageReduceSize()
        msg.contract_id = contract_id
        msg.entry_id = entry_id
        msg.size_decrement = size_decrement
        return self.__submit(msg)

    def get_contract(self, contract_id, all_contracts=False):
        """\
        Get contract details for a particular contract ID or if `all_contracts`
        is set then get details for all contracts on the exchange.

        :param contract_id: The contract ID.
        :param all_contracts: Whether to get details for all contracts (False).
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageGetContractDetail()
        msg.contract_id = contract_id
        msg.all_contracts = all_contracts
        return self.__submit(msg)

    def get_book_state(self, contract_id):
        """\
        Get the book state for a particular contract ID.

        :param contract_id: The contract ID.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageGetBookState()
        msg.contract_id = contract_id
        return self.__submit(msg)

    def replay(self, begin, end):
        """\
        Replay action reports between ``begin`` and ``end`` timestamps.

        :param begin: A timestamp in seconds since the Epoch.
        :param end: A timestamp in seconds since the Epoch.
        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        msg = api.MessageReplay()
        msg.begin_timestamp = begin
        msg.end_timestamp = end
        return self.__submit(msg)

    def __make_submission(self, msg, is_ask=False, size=0, price_in_cents=0,
            mox2=False, scp_sweeper_mpid=0, scp_sweeper_cid=0, **kwargs):
        """\
        Make a submission message with default parameters.
        """
        msg.is_ask = is_ask
        msg.size = size
        msg.price_in_cents = price_in_cents
        try:
            msg.scp_sweeper_mpid = scp_sweeper_mpid
            msg.scp_sweeper_cid = scp_sweeper_cid
            msg.mox2 = mox2
        except AttributeError:
            # XXX: Swallow attribute errors as these fields can only be set on
            # messages that support them.
            pass

        # Set any other arbitrary fields requested
        for k, v in kwargs.items():
            setattr(msg, k, v)
        return msg

    def __submit(self, msg):
        """\
        Submit a protocol message to the exchange via :class:`ledgerx.context.Context`.

        :returns: The return value of :method:`ledgerx.context.Context.submit`.
        """
        if not self.context.is_alive():
            raise LXError('must start Context first')
        return self.context.submit(msg)

