# Copyright 2014 NYBX Inc.
# All rights reserved.

"""
:module: ledgerx.context
:synopsis: A module to manage the context of a ledgerx exchange client.
:author: Amr Ali <amr@ledgerx.com>
"""

import zmq

from threading import Thread, Event
from concurrent import futures
from zmq.utils.monitor import parse_monitor_message

from ledgerx import settings
from ledgerx.exceptions import LXError
from ledgerx.api.client import protocol as api
from ledgerx.mixin import public
from ledgerx.store import IDStoreItem
from ledgerx.utils import logger
from ledgerx.utils.meta import KindSingletonMeta
from ledgerx.protocol.sockets import (dealer_socket,
        router_socket, sub_socket, secure_socket, pair_socket)

@public
class Context(object, metaclass=KindSingletonMeta):
    """\
    A ledgerx exchange client context.
    """

    def __init__(self, private_key, public_key, server_key, processor_class):
        self.context = zmq.Context.instance(settings.THREAD_COUNT)
        self.running = Event()
        self.io_loop = Thread(target=self._run_device)
        self.cmd_endpoint = "inproc://Context"
        self.state_endpoint = "inproc://Context.state"
        self.private_key = private_key
        self.public_key = public_key
        self.server_key = server_key
        self.proc_class = processor_class

    def start(self):
        """\
        Start ledgerx context I/O loop.
        """
        if self.running.is_set():
            raise LXError("context already started")

        self.running.set()
        self.io_loop.start()

    def stop(self):
        """\
        Stop LedgerX context I/O loop.
        """
        if not self.running.is_set():
            raise LXError("context already stopped")

        self.running.clear()
        cmd_sock = dealer_socket(self.context)
        cmd_sock.connect(self.cmd_endpoint)
        cmd_sock.send_pyobj('EXIT')

    def join(self, timeout=None):
        """\
        Join LedgerX context I/O loop.

        :param timeout: Timeout in seconds.
        """
        self.io_loop.join(timeout=timeout)
        self.context.term()

    def is_alive(self):
        """\
        Check whether context's I/O loop is engaged.
        """
        return self.io_loop.is_alive() and self.running.is_set()

    def submit(self, msg):
        """\
        Submit a protocol message to the exchange.

        :param msg: The message to be sent to the exchange.
        :returns: A reference of :class:`ledgerx.store.IDStoreItem`.
        """
        payload = msg.dumps()
        cmd_sock = dealer_socket(self.context)
        cmd_sock.connect(self.cmd_endpoint)
        cmd_sock.send_pyobj(['SEND', payload])
        cmd_sock.close()
        settings.ID_STORE[msg.mid] = IDStoreItem(msg)
        return settings.ID_STORE[msg.mid]

    def _run_device(self):
        try:
            self.jobs = []
            self.thpool = futures.ThreadPoolExecutor(max_workers=1)
            self.client_sock = secure_socket(dealer_socket,
                    self.private_key, self.public_key, self.server_key,
                    self.context)
            self.client_sock.monitor(self.state_endpoint)
            self.client_sock.connect(settings.EXCHANGE_ENDPOINT)

            self.client_state_sock = pair_socket(self.context)
            self.client_state_sock.connect(self.state_endpoint)

            self.publisher_sock = secure_socket(sub_socket,
                    self.private_key, self.public_key, self.server_key,
                    self.context)
            self.publisher_sock.connect(settings.EXCHANGE_PUBLISHER_ENDPOINT)
            self.publisher_sock.subscribe = b''

            self.cmd_sock = router_socket(self.context)
            self.cmd_sock.bind(self.cmd_endpoint)

            self.poller = zmq.Poller()
            self.poller.register(self.client_sock, zmq.POLLIN)
            self.poller.register(self.client_state_sock, zmq.POLLIN)
            self.poller.register(self.publisher_sock, zmq.POLLIN)
            self.poller.register(self.cmd_sock, zmq.POLLIN)

            # Context's I/O loop
            self.__io_loop()

        except:
            logger.exception("error occurred in context I/O loop")

        finally:
            self.client_sock.close()
            self.client_state_sock.close()
            self.publisher_sock.close()
            self.cmd_sock.close()
            self.thpool.shutdown()

    def __io_loop(self):
        processor = self.proc_class()

        while self.running.is_set():
            socks = dict(self.poller.poll())

            # Received a message on exchange's socket
            if self.client_sock in socks:
                msg = self.client_sock.recv()
                obj = api.MessageParser.parse(msg)
                self.__submit_thpool(processor.handle, obj)

            # Received an exchange's socket state event
            if self.client_state_sock in socks:
                msg = self.client_state_sock.recv_multipart()
                state = parse_monitor_message(msg)
                self.__submit_thpool(processor.handle_state_event, state)

            # Received a message from exchange's publisher network
            if self.publisher_sock in socks:
                topic, msg = self.publisher_sock.recv_multipart()
                obj = api.MessageParser.parse(msg)
                self.__submit_thpool(processor.handle, obj)

            # Received a message on command socket
            if self.cmd_sock in socks:
                sid = self.cmd_sock.recv()
                msg = self.cmd_sock.recv_pyobj()

                # Handle exit interrupt
                if msg == 'EXIT':
                    self.running.clear()

                # Handle other messages to send to the exchange
                elif isinstance(msg, list) and msg[0] == 'SEND':
                    self.client_sock.send(msg[1])

    def __submit_thpool(self, func, *args, **kwargs):
        self.jobs.append(
                self.thpool.submit(func, *args, **kwargs))
        # TODO: Determine if internal futures queue of jobs need to be emptied.
        # We don't require to wait on jobs at all, we don't care about their
        # return values here.
        #
        #futures.wait(self.jobs, timeout=1e-9,
        #        return_when=futures.FIRST_COMPLETED)

