# Copyright 2014 NYBX Inc.
# See LICENSE file for details.

"""
:module: ledgerx.processor
:synopsis: A module to provide message processing facilities.
:author: Amr Ali <amr@ledgerx.com>
"""

import abc
import zmq

from ledgerx import settings
from ledgerx.exceptions import LXError
from ledgerx.api import client as api
from ledgerx.mixin import public
from ledgerx.store import IDStoreItem
from ledgerx.utils import logger

@public
class MessageProcessor(object, metaclass=abc.ABCMeta):
    """\
    An abstract message processor class to provide hooks for processing
    different types of messages.

    A child of this class will have to provide implementation for all abstract
    methods to handle different types of messages that come from the exchange.
    """

    def handle(self, msg):
        """\
        Handle a message and call the appropriate method.

        :param msg: A parsed message object.
        :returns: Whether the call was successful.
        """
        try:
            handler = getattr(self,
                    'handle_v' + msg.mversion.replace('.', '_'), None)
            if not handler:
                raise LXError("unimplemented version processor")

            prev_item = IDStoreItem()
            if msg.mid in settings.ID_STORE:
                prev_item = settings.ID_STORE[msg.mid]
            settings.ID_STORE[msg.mid] = IDStoreItem(msg, prev_item.callback)

            try:
                prev_item.callback(msg, prev_item.msg)
            except:
                logger.exception("error occurred in callback")

            return handler(msg, prev_item.msg)
        except:
            logger.exception("error occurred in message processor")
            return False

    def handle_state_event(self, state):
        """\
        Handle a state event and call the appropriate method.

        :param state: The resulted dictionary from calling
            :func:`zmq.utils.monitor.parse_monitor_message`.
        :returns: Whether the call was successful.
        """
        event, endpoint = state['event'], state['endpoint']
        try:
            if event == zmq.EVENT_CONNECTED:
                return self.handle_event_connected(endpoint)
            elif event == zmq.EVENT_CONNECT_RETRIED:
                return self.handle_event_retried(endpoint, state['value'])
            elif event == zmq.EVENT_CLOSED:
                return self.handle_event_closed(endpoint)
            elif event == zmq.EVENT_MONITOR_STOPPED:
                return self.handle_event_monitor_stopped(endpoint)
            else:
                return True # XXX: We are not interested in the other events
        except:
            logger.exception("error occurred in message processor")
            return False

    def handle_v0_0_1(self, msg, prev_msg):
        """\
        Handle a protocol message v0.0.1 and call the appropriate method.

        :param msg: A v0.0.1 protocol message.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        :raises: :class:`ledgerx.LXError`
        """
        if isinstance(msg, api.v0_0_1.MessageStatus):
            return self.handle_status_message(msg, prev_msg)
        elif isinstance(msg, api.v0_0_1.MessageActionReport):
            return self.handle_action_report(msg, prev_msg)
        elif isinstance(msg, api.v0_0_1.MessageContractDetail):
            return self.handle_contract_detail(msg, prev_msg)
        elif isinstance(msg, api.v0_0_1.MessageBookTop):
            return self.handle_book_top(msg, prev_msg)
        elif isinstance(msg, api.v0_0_1.MessageBookState):
            return self.handle_book_state(msg, prev_msg)
        raise LXError("unimplemented message processor")

    @abc.abstractmethod
    def handle_action_report(self, msg, prev_msg):
        """\
        An abstract method to be overridden by a subclass to provide a handler
        for action report messages.

        :param msg: A parsed message object.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        """
        pass

    @abc.abstractmethod
    def handle_contract_detail(self, msg, prev_msg):
        """\
        An abstract method to be overridden by a subclass to provide a handler
        for contract detail messages.

        :param msg: A parsed message object.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        """
        pass

    @abc.abstractmethod
    def handle_status_message(self, msg, prev_msg):
        """\
        An abstract method to be overridden by a subclass to provide a handler
        for status messages.

        :param msg: A parsed message object.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        """
        pass

    @abc.abstractmethod
    def handle_book_top(self, msg, prev_msg):
        """\
        An abstract method to be overridden by a subclass to provide a handler
        for book top messages.

        :param msg: A parsed message object.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        """
        pass

    @abc.abstractmethod
    def handle_book_state(self, msg, prev_msg):
        """\
        An abstract method to be overridden by a subclass to provide a handler
        for book state messages.

        :param msg: A parsed message object.
        :param prev_msg: The previous message with the same MID if it exists.
            None otherwise.
        :returns: Whether the call was successful.
        """
        pass

    def handle_event_connected(self, endpoint):
        """\
        A method to be overridden by a subclass to provide a handler
        for a connected state event.

        :param endpoint: The endpoint configured on the socket.
        :returns: Whether the call was successful.
        """
        return True

    def handle_event_retried(self, endpoint, interval):
        """\
        A method to be overridden by a subclass to provide a handler
        for a connection retried state event.

        :param endpoint: The endpoint configured on the socket.
        :param interval: The reconnect interval in milliseconds.
        :returns: Whether the call was successful.
        """
        return True

    def handle_event_closed(self, endpoint):
        """\
        A method to be overridden by a subclass to provide a handler
        for a socket closed state event.

        :param endpoint: The endpoint configured on the socket.
        :returns: Whether the call was successful.
        """
        return True

    def handle_event_monitor_stopped(self, endpoint):
        """\
        A method to be overridden by a subclass to provide a handler
        for a socket unexpectedly disconnected state event.

        :param endpoint: The endpoint configured on the socket.
        :returns: Whether the call was successful.
        """
        return True

