# Copyright 2014 NYBX Inc.
# All rights reserved.

"""
:module: ledgerx.store
:synopsis: Memory store for MIDs.
:author: Amr Ali <amr@ledgerx.com>
"""

from collections import MutableMapping
from multiprocessing import Manager
from ledgerx.mixin import public

@public
class MemoryStore(MutableMapping):
    """\
    An inter-process in-memory store to provide common facilities for loading
    data from different sources. Namely from another process. This container
    can be constructed to not be shared among forked processes.
    """

    def __init__(self, shared=False):
        if shared:
            self.manager = Manager()
            self._l = self.manager.list()
        else:
            self._l = []
        self._l.append({})

    def __getitem__(self, key):
        return self._l[0].__getitem__(key)

    def __setitem__(self, key, value):
        d = self._l[0]
        d.__setitem__(key, value)
        self._l[0] = d

    def __delitem__(self, key):
        d = self._l[0]
        d.__delitem__(key)
        self._l[0] = d

    def __contains__(self, key):
        return self._l[0].__contains__(key)

    def __iter__(self):
        return self._l[0].__iter__()

    def __len__(self):
        return self._l[0].__len__()

@public
class IDStoreItem(object):
    """\
    A class to represent a MessageIDStore's item.

    :param msg: A property that holds a message object.
    :param callback: A callable that accepts two arguments, the first is
        the just received message object and the second is the previous message
        with the same message ID.
    """

    def __init__(self, msg=None, cb=lambda x, y: None):
        self._msg = msg
        self._cb = cb

    @property
    def msg(self):
        return self._msg

    @msg.setter
    def msg(self, val):
        self._msg = val

    @property
    def callback(self):
        return self._cb

    @callback.setter
    def callback(self, val):
        if not callable(val):
            raise ValueError("value must be a callable")
        self._cb = val

@public
class MessageIDStore(dict):
    """\
    A store to keep track of the IDs of in-progress messages.
    """

    def __setitem__(self, key, value):
        """\
        Check whether value is of type :class:`IDStoreItem`.
        """
        if not isinstance(value, IDStoreItem):
            raise ValueError("value must be of type IDStoreItem")
        super().__setitem__(key, value)

