# Copyright 2014 NYBX Inc.
# All rights reserved.

"""
:module: ledgerx.test.common.context
:synopsis: A module for shared functionality of :class:`ledgerx.context.Context`.
:author: Amr Ali <amr@ledgerx.com>
"""

import zmq
import zmq.auth
import zmq.auth.thread
import unittest

from collections import MutableMapping

from ledgerx import settings
from ledgerx.context import Context
from ledgerx.protocol.crypto import KeyPair
from ledgerx.protocol.sockets import (secure_socket,
        dealer_socket, router_socket, pub_socket)

class Authenticator(zmq.auth.Authenticator):
    """\
    Wrapper for :class:`zmq.auth.Authenticator` to override configure_curve.
    """

    def configure_curve(self, store=None):
        """\
        Override configure_curve to allow for custom key stores.
        """
        if not store:
            self.allow_any = True
        else:
            self.allow_any = False
            if isinstance(store, MutableMapping):
                self.certs = store
            else:
                self.log.error("Invalid CURVE certs store: %s", type(store))

class AuthenticationThread(zmq.auth.thread.AuthenticationThread):
    """\
    Wrapper for :class:`zmq.auth.thread.AuthenticationThread` to override
    authenticator class type.
    """

    def __init__(self, context, endpoint, encoding='utf8', log=None):
        super(self.__class__, self).__init__(context, endpoint, encoding, log)
        self.authenticator = Authenticator(context, encoding=encoding, log=log)

class ThreadAuthenticator(zmq.auth.thread.ThreadAuthenticator):
    """\
    Wrapper for :class:`zmq.auth.thread.ThreadAuthenticator` to override
    configure_curve.
    """

    def configure_curve(self, store=None):
        """\
        Configure CURVE authentication key store. CURVE authentication
        uses a dict that holds all public client certificates, i.e. their
        public keys. The certificates must be in zcert_save () format.
        You can add and remove certificates in that store at any time.
        To allow all client keys without checking, specify `None` for the `store`.
        """
        if not self.thread:
            raise RuntimeError('Must start the authenticator thread first')
        self.thread.authenticator.configure_curve(store)

    def start(self):
        """\
        Override :method:`zmq.auth.thread.ThreadAuthenticator.start` to use
        our version of the :class:`AuthenticationThread`.
        """
        self.pipe = self.context.socket(zmq.PAIR)
        self.pipe.linger = 1
        self.pipe.bind(self.pipe_endpoint)
        self.thread = AuthenticationThread(self.context, self.pipe_endpoint,
                encoding=self.encoding, log=self.log)
        self.thread.start()

def generate_session_keypairs():
    """\
    Generate a server and a client key pairs.
    """
    return (KeyPair.generate(), KeyPair.generate())

def generate_server_sockets(keypair, context):
    """\
    Generate a server and a publisher secure sockets.
    """
    server_sock = secure_socket(router_socket,
            keypair.private, keypair.public, ctx=context)
    server_sock.bind(settings.EXCHANGE_ENDPOINT)

    publisher_sock = secure_socket(pub_socket,
            keypair.private, keypair.public, ctx=context)
    publisher_sock.bind(settings.EXCHANGE_PUBLISHER_ENDPOINT)

    return (server_sock, publisher_sock)

def generate_context_test_mock(msg_proc_klass):
    """\
    A function to generate a parent :class:`unittest.TestCase` for other
    tests to inherit from to provide basic mocking of the exchange's interface.
    """
    class __MockExchange(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            server_keypair, client_keypair = generate_session_keypairs()
            cert_store = {'*': [client_keypair.public]}

            # Prepare context

            cls.context = Context.new(client_keypair.private,
                    client_keypair.public, server_keypair.public,
                    msg_proc_klass)
            cls.ctx = zmq.Context.instance()

            # Start a ZAP handler

            cls.auth = ThreadAuthenticator(cls.ctx)
            cls.auth.start()
            cls.auth.configure_curve(cert_store)

            # Mock exchange's server sockets

            cls.server_socket, cls.publisher_socket = generate_server_sockets(
                    server_keypair, cls.ctx)

            # Fire up the context

            cls.context.start()

            # Establish comm with context's command socket

            cls.cmd_socket = dealer_socket(cls.ctx)
            cls.cmd_socket.connect(cls.context.cmd_endpoint)

        @classmethod
        def tearDownClass(cls):
            cls.publisher_socket.close()
            cls.server_socket.close()
            cls.cmd_socket.close()
            cls.auth.stop()
            cls.context.stop()
            cls.context.join(timeout=1)
            cls.ctx.term()

    return __MockExchange

