# Copyright 2014 NYBX Inc.
# All rights reserved.

"""
:module: ledgerx.utils.meta
:synopsis: Module for metaclass patterns.
:author: Amr Ali <amr@ledgerx.com>
"""

from abc import ABCMeta
from collections import defaultdict

class SingletonAbstractMeta(ABCMeta):
    """\
    A Singleton pattern abstract metaclass.
    """

    def __init__(cls, name, bases, attrs):
        super(SingletonAbstractMeta, cls).__init__(name, bases, attrs)
        setattr(cls, 'new', super(SingletonAbstractMeta, cls).__call__)
        cls._instance = defaultdict(str)

    def __call__(cls, *args, **kwargs):
        key = cls._make_key(*args, **kwargs)
        if key not in cls._instance:
            cls._instance[key] = super(SingletonAbstractMeta, cls).__call__(*args, **kwargs)
        return cls._instance[key]

    @classmethod
    def _make_key(cls, *args, **kwargs):
        """A function responsible for generating a key to check against and see
        if an instance was already constructed for that key.

        This is an abstract function, it is required to override it.
        """
        raise NotImplementedError()

class KindSingletonMeta(SingletonAbstractMeta):
    """\
    A singleton pattern metaclass that uses the underlying class' name
    as the unique constraint.
    """

    @classmethod
    def _make_key(cls, *args, **kwargs):
        return cls.__class__

